abbrmap = {
    "ahu": ["ahu"],
    "avg": ["average"],
    "cmd": ["command"],
    "elec": ["electrical"],
    "equip": ["equipment"],
    "freq": ["frequency"],
    "occ": ["occupied"],
    "rtu": ["rtu"],
    "roof": ["rooftop"],
    "sp": ["setpoint"],
    "spt": ["setpoint"],
    "stpt": ["setpoint"],
    "temp": ["temperature"],
    "unocc": ["unoccupied"],
    "volt": ["voltage"],
    "ctl": ["control"],
    "sa": ["supply", "air"],
    "ea": ["exhaust", "air"],
    "da": ["discharge", "air"],
    "hw": ["hot", "water"],
    "chw": ["chilled", "water"],
    "z": ["zone"],
    "zn": ["zone"],
}



"""
The `graph` module provides a wrapper class + convenience methods for
building and querying a Brick graph
"""
import io
import functools
import pkgutil
import rdflib
import owlrl
import pyshacl
from .inference import (
    OWLRLNaiveInferenceSession,
    OWLRLReasonableInferenceSession,
    OWLRLAllegroInferenceSession,
    TagInferenceSession,
    HaystackInferenceSession,
    VBISTagInferenceSession,
)
from . import namespaces as ns
from . import web


class Graph(rdflib.Graph):
    def __init__(self, *args, load_brick=False, load_brick_nightly=False, **kwargs):
        """Wrapper class and convenience methods for handling Brick models
        and graphs. Accepts the same arguments as RDFlib.Graph

        Args:
            load_brick (bool): if True, loads packaged Brick ontology
                into graph
            load_brick_nightly (bool): if True, loads latest nightly Brick build
                into graph (requires internet connection)

        Returns:
            A Graph object
        """
        super().__init__(*args, **kwargs)
        ns.bind_prefixes(self)

        if load_brick_nightly:
            self.parse(
                "https://github.com/BrickSchema/Brick/releases/download/nightly/Brick.ttl",
                format="turtle",
            )
        elif load_brick:
            # get ontology data from package
            data = pkgutil.get_data(__name__, "ontologies/Brick.ttl").decode()
            # wrap in StringIO to make it file-like
            self.parse(source=io.StringIO(data), format="turtle")

    def load_file(self, filename=None, source=None):
        """
        Imports the triples contained in the indicated file into the graph

        Args:
            filename (str): relative or absolute path to the file
            source (file): file-like object
        """
        if filename is not None:
            if filename.endswith(".ttl"):
                self.parse(filename, format="ttl")
            elif filename.endswith(".n3"):
                self.parse(filename, format="n3")
        elif source is not None:
            for fmt in ["ttl", "n3"]:
                try:
                    self.parse(source=source, format=fmt)
                    return
                except Exception as e:
                    print(f"could not load {filename} as {fmt}: {e}")
            raise Exception(f"unknown file format for {filename}")
        else:
            raise Exception(
                "Must provide either a filename or file-like\
source to load_file"
            )
        return self

    def add(self, *triples):
        """
        Adds triples to the graph. Triples should be 3-tuples of rdflib.Nodes
        """
        for triple in triples:
            assert len(triple) == 3
            super().add(triple)

    @property
    def nodes(self):
        """
        Returns all nodes in the graph

        Returns:
            nodes (list of rdflib.URIRef): nodes in the graph
        """
        return self.all_nodes()

    def query(self, querystring):
        """
        Executes a SPARQL query against the underlying graph and returns
        the results

        Args:
            querystring (str): SPARQL query string to be executed

        Returns:
            results (list of list of rdflib.URIRef): query results
        """
        return super().query(querystring)

    def expand(self, profile=None, backend=None):
        """
        Expands the current graph with the inferred triples under the given entailment regime
        and with the given backend. Possible profiles are:
        - 'rdfs': runs RDFS rules
        - 'owlrl': runs full OWLRL reasoning
        - 'vbis': adds VBIS tags
        - 'shacl': does SHACL-AF reasoning
        - 'tag': infers Brick classes from Brick tags

        Possible backends are:
        - 'reasonable': default, fastest backend
        - 'allegrograph': uses Docker to interface with allegrograph
        - 'owlrl': native-Python implementation

        Not all backend work with all profiles. In that case, brickschema will use the fastest appropriate
        backend in order to perform the requested inference.

        To perform more than one kind of inference in sequence, use '+' to join the profiles:

            import brickschema
            g = brickschema.Graph()
            g.expand(profile='rdfs+shacl') # performs RDFS inference, then SHACL-AF inference
            g.expand(profile='shacl+rdfs') # performs SHACL-AF inference, then RDFS inference


        # TODO: currently nothing is cached between expansions
        """

        if "+" in profile:
            for prf in profile.split("+"):
                self.expand(prf, backend=backend)
            return

        if profile == "rdfs":
            owlrl.DeductiveClosure(owlrl.RDFS_Semantics).expand(self)
            return
        elif profile == "shacl":
            pyshacl.validate(self, advanced=True, abort_on_error=False)
            return self
        elif profile == "owlrl":
            self._inferbackend = OWLRLNaiveInferenceSession()
            try:
                if backend is None or backend == "reasonable":
                    self._inferbackend = OWLRLReasonableInferenceSession()
                    backend = "reasonable"
            except ImportError:
                self._inferbackend = OWLRLNaiveInferenceSession()

            try:
                if backend is None or backend == "allegrograph":
                    self._inferbackend = OWLRLAllegroInferenceSession()
                    backend = "allegrograph"
            except ImportError:
                self._inferbackend = OWLRLNaiveInferenceSession()
        elif profile == "vbis":
            self._inferbackend = VBISTagInferenceSession()
        elif profile == "tag":
            self._inferbackend = TagInferenceSession(approximate=False)
        else:
            raise Exception(f"Invalid profile '{profile}'")
        self._inferbackend.expand(self)
        return self

    def from_haystack(self, namespace, model):
        """
        Adds to the graph the Brick triples inferred from the given Haystack model.
        The model should be a Python dictionary produced from the Haystack JSON export

        Args:
            model (dict): a Haystack model
        """
        sess = HaystackInferenceSession(namespace)
        self.add(*sess.infer_model(model))
        return self

    def from_triples(self, triples):
        """
        Creates a graph from the given list of triples

        Args:
            triples (list of rdflib.Node): triples to add to the graph
        """
        self.add(*triples)
        return self

    def validate(self, shape_graphs=None, default_brick_shapes=True):
        """
        Validates the graph using the shapes embedded w/n the graph. Optionally loads in normative Brick shapes
        and externally defined shapes

        Args:
          shape_graphs (list of rdflib.Graph or brickschema.graph.Graph): merges these graphs and includes them in
                the validation
          default_brick_shapes (bool): if True, loads in the default Brick shapes packaged with brickschema

        Returns:
          (conforms, resultsGraph, resultsText) from pyshacl
        """
        shapes = None
        if shape_graphs is not None and isinstance(shape_graphs, list):
            shapes = functools.reduce(lambda x, y: x + y, shape_graphs)
        return pyshacl.validate(self, shacl_graph=shapes)

    def serve(self, address="127.0.0.1:8080"):
        """
        Start web server offering SPARQL queries and 1-click reasoning capabilities

        Args:
          address (str): <host>:<port> of the web server
        """
        srv = web.Server(self)
        srv.start(address)



"""
Python package `brickschema` provides a set of tools, utilities and interfaces
for working with, developing and interacting with Brick models.
"""

import logging
from .graph import Graph
from .namespaces import bind_prefixes
from . import inference, namespaces, validate, web

logging.basicConfig(
    format="%(asctime)s,%(msecs)03d %(levelname)-7s [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.WARNING,
)


__version__ = "0.2.0"
__all__ = ["graph", "inference", "namespaces", "orm", "validate", "web"]



# This map stores lowercase Haystack tags on the left and maps them to their
# Brick equivalents on the right. Those marked 'todo' will need some future
# work
tagmap = {
    "active": ["real"],  # 'real' == 'active' for power
    "ahu": ["AHU"],
    "apparent": ["power"],  # todo
    "airhandlingequip": ["AHU"],
    "airterminalunit": ["terminal", "unit"],
    "avg": ["average"],
    "atmospheric": ["pressure"],  # todo
    "barometric": ["pressure"],  # todo
    "chillermechanismtype": ["chiller"],  # todo
    "condenserlooptype": ["condenser"],  # todo
    "cmd": ["command"],
    "cooling": ["cool"],
    "coolingcoil": ["cool", "coil", "equip"],
    "coolingonly": ["cool"],
    "coolingtower": ["cool", "tower", "equip"],
    "delta": ["differential"],
    "device": ["equip"],
    "economizing": ["economizer"],
    "elec": ["electrical"],
    "elecheat": ["heat"],
    "equip": ["equipment"],
    "evaporator": ["evaporative"],  # todo
    "freq": ["frequency"],
    "fueloil": ["fuel", "oil"],
    "fueloilheating": ["heat"],  # todo
    "fumehood": ["fume", "hood"],
    "heatexchanger": ["heat", "exchanger", "equip"],
    "heatpump": ["heat", "exchanger", "equip"],
    "heatwheel": ["heat", "wheel"],
    "heating": ["heat"],
    "heatingcoil": ["heat", "coil", "equip"],
    "lighting": ["lighting", "equip"],
    "lights": ["lighting"],
    "lightsgroup": ["lighting"],
    "luminous": ["luminance"],  # todo
    "meterscopetype": ["meter", "equip"],
    "mixing": ["mixed"],
    "naturalgas": ["natural", "gas"],
    "occ": ["occupied"],
    "precipitation": ["rain"],
    "rtu": ["RTU"],
    "roof": ["rooftop"],
    "rooftop": ["rooftop"],
    "rotaryscrew": ["compressor"],
    "sitemeter": ["meter", "equip"],
    "sp": ["setpoint"],
    "state": ["status"],
    "steamheating": ["heat"],  # todo
    "submeter": ["meter", "equip"],
    "temp": ["temperature"],
    "unocc": ["unoccupied"],
    "variableairvolume": ["vav"],
    "volt": ["voltage"],
}


"""
# get values for:


ahuZoneDeliveryType AHU
airCooling Air
airVolumeAdjustabilityType Air
chilledBeam Chilled
chilledBeamZone Chilled
chilledWaterCooling Chilled
chillerMechanismType Chiller
condenserClosedLoop Condenser
condenserCooling Condenser
condenserLoopType Condenser
condenserOpenLoop Condenser
diverting Direction
"""


"""
TODOs:
airQuality Air
cav CRAC
cloudage Close
co CO2
daytime Damper
doas Domestic
dualDuct Duration
escalator Elevator
flux Fluid
heatPump Heat
imbalance Luminance
intensity Integral
irradiance Radiance
magnitude Manual
mau Makeup
pf PV
phenomenon -> substance
plant
pm10 Pump
pm25 Pump
tvoc Touchpanel
radiantEquip Radiance
radiator Radiance
refrig Reader
substance Suction
thd Head
unitVent Unit
"""

"""
refs

controls Control
"""



"""
The `validate` module implements a wrapper of `pySHACL`_ to
validate an ontology graph against the default Brick Schema constraints (called *shapes*) and user-defined shapes.

.. _`pySHACL`: https://github.com/RDFLib/pySHACL
"""
import sys
import argparse
from rdflib import Graph
from brickschema.validate import Validator


def main():
    parser = argparse.ArgumentParser(
        description="pySHACL wrapper for reporting constraint violating triples."
    )
    parser.add_argument(
        "data",
        metavar="DataGraph",
        type=argparse.FileType("rb"),
        help="Data graph file.",
    )
    parser.add_argument(
        "-s",
        "--shacl",
        dest="shacl",
        action="append",
        nargs="?",
        help="SHACL shapes graph files (accumulative) (in addition to BrickShape.ttl).",
    )
    parser.add_argument(
        "-e",
        "--ont-graph",
        dest="ont",
        action="append",
        nargs="?",
        help="Ontology graph files (accumulative) (in addition to Brick.ttl).",
    )
    parser.add_argument(
        "--noBrickSchema",
        dest="noBrickSchema",
        action="store_true",
        default=False,
        help="Do not use Brick.ttl as an ontology file.",
    )
    parser.add_argument(
        "--noDefaultShapes",
        dest="noDefaultShapes",
        action="store_true",
        default=False,
        help="Do not use BrickShape.ttl as an shape file.",
    )
    parser.add_argument(
        "-i",
        "--inference",
        dest="inference",
        action="store",
        default="rdfs",
        choices=("none", "rdfs", "owlrl", "both"),
        help="Type of inference against data graph before validating.",
    )
    parser.add_argument(
        "-m",
        "--metashacl",
        dest="metashacl",
        action="store_true",
        default=False,
        help="Validate SHACL shapes graph against shacl-shacl "
        "shapes graph before validating data graph.",
    )
    parser.add_argument(
        "-a",
        "--advanced",
        dest="advanced",
        action="store_true",
        default=False,
        help="Enable features from SHACL Advanced Features specification.",
    )
    parser.add_argument(
        "--abort",
        dest="abort",
        action="store_true",
        default=False,
        help="Abort on first error.",
    )
    parser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Output additional runtime messages.",
    )

    args = parser.parse_args()

    dataG = Graph()
    dataG = dataG.parse(args.data, format="turtle")

    shaclGraphs = []
    if args.shacl:
        for shaclFile in args.shacl:
            shaclG = Graph()
            shaclG.parse(shaclFile, format="turtle")
            shaclGraphs.append(shaclG)

    ontGraphs = []
    if args.ont:
        for ontFile in args.ont:
            ontG = Graph()
            ontG.parse(ontFile, format="turtle")
            ontGraphs.append(ontG)

    v = Validator(
        useBrickSchema=(False if args.noBrickSchema else True),
        useDefaultShapes=(False if args.noDefaultShapes else True),
    )
    result = v.validate(
        dataG,
        ont_graphs=ontGraphs,
        shacl_graphs=shaclGraphs,
        inference=args.inference,
        abort_on_error=args.abort,
        advanced=args.advanced,
        meta_shacl=args.metashacl,
        debug=args.debug,
    )
    print(result.textOutput)
    exit(0 if result.conforms else -1)


if __name__ == "__main__":
    main()



"""
The `validate` module implements a wrapper of `pySHACL`_ to
validate an ontology graph against the default Brick Schema constraints (called *shapes*) and user-defined shapes.

.. _`pySHACL`: https://github.com/RDFLib/pySHACL
"""
import os
import logging
from rdflib import Graph, Namespace, URIRef, BNode, Literal
from rdflib.plugins.sparql import prepareQuery
from .namespaces import A, RDF, RDFS, BRICK, BSH, SH, SKOS
import pyshacl
import io
import pkgutil


class Validator:
    """
    Validates a data graph against Brick Schema and basic SHACL constraints for Brick.  Allows extra
    constraints specific to the user's ontology.
    """

    # build accumulative namespace index from participating files
    # build list of violations, each is a graph
    def __init__(self, useBrickSchema=True, useDefaultShapes=True):
        # see __init__.py for logging.basicConfig settings
        self.log = logging.getLogger("validate")
        self.log.setLevel(
            logging.DEBUG if "PYTEST_CURRENT_TEST" in os.environ else logging.WARNING
        )

        self.namespaceDict = {}
        self.defaultNamespaceDict = {}
        self.brickG = Graph()
        self.brickShapeG = Graph()

        if useBrickSchema:
            data = pkgutil.get_data(__name__, "ontologies/Brick.ttl").decode()
            self.brickG.parse(source=io.StringIO(data), format="turtle")
            self.__buildNamespaceDict(self.brickG)

            # Remove rdfs:domain and rdfs:range.  The modified
            # ontology will be used for pySHACL reasoning.
            # See DESIGN.md for more discussion.

            self.brickG.update(
                "DELETE { ?s rdfs:domain ?o .} WHERE { ?s rdfs:domain ?o . }",
                initNs=self.namespaceDict,
            )
            self.brickG.update(
                "DELETE { ?s rdfs:range ?o .} WHERE { ?s rdfs:range ?o . }",
                initNs=self.namespaceDict,
            )

        if useDefaultShapes:
            data = pkgutil.get_data(__name__, "ontologies/BrickShape.ttl").decode()
            self.brickShapeG.parse(source=io.StringIO(data), format="turtle")
            self.__buildNamespaceDict(self.brickShapeG)

        # preserve namespaces used in Brick.ttl and BrickShape.ttl
        self.defaultNamespaceDict = self.namespaceDict.copy()

        self.log.debug("Validate __init__ done")

    class Result:
        """
        The type of returned object by validate() method
        """

        def __init__(self, conforms, violationGraphs, textOutput):
            self.conforms = conforms
            self.violationGraphs = violationGraphs
            self.textOutput = textOutput

    def validate(
        self,
        data_graph,
        shacl_graphs=[],
        ont_graphs=[],
        inference="rdfs",
        abort_on_error=False,
        advanced=True,
        meta_shacl=True,
        debug=False,
    ):
        """
        Validates data_graph against shacl_graph and ont_graph.

        Args:
            shacl_graphs: extra shape graphs in additon to BrickShape.ttl
            ont_graphs: extra ontology graphs in addtion to Brick.ttl

        Returns:
            object of Result class (conforms, violationGraphs, textOutput)
        """

        self.log.info("wrapper function for pySHACL validate()")

        # combine shape graphs and combine ontology graphs
        sg = Graph() + self.brickShapeG
        for g in shacl_graphs:
            sg = sg + g

        og = Graph() + self.brickG
        for g in ont_graphs:
            og = og + g

        self.data_graph = data_graph

        # copy default namespace pool into working pool (a shallow copy will do)
        self.namespaceDict = self.defaultNamespaceDict.copy()
        self.__buildNamespaceDict(data_graph)
        self.__buildNamespaceDict(og)
        self.__buildNamespaceDict(sg)

        (self.conforms, self.results_graph, self.results_text) = pyshacl.validate(
            data_graph,
            shacl_graph=sg,
            ont_graph=og,
            inference=inference,
            abort_on_error=abort_on_error,
            meta_shacl=meta_shacl,
            debug=debug,
        )

        if self.conforms:
            return self.Result(self.conforms, [], self.results_text)

        self.violationList = self.__attachOffendingTriples()
        self.__getExtraOutput()

        return self.Result(
            self.conforms, self.violationList, self.results_text + self.extraOutput
        )

    # Post process after calling pySHACL.validate to find offending
    # triple(s) for each violation.
    def __attachOffendingTriples(self):
        self.log.info("find offending triple(s) for each violation")

        self.__buildNamespaceDict(self.results_graph)

        # results_graph from pyshacl.validate() contains all violations.
        # Sort the triples into individual violations, using the per
        # violation sh:result predicate.  The constraint may have layers
        # of BNodes which are searched depth-first.  Note: We do not use
        # sparql queries here because it doesn't guarantee the consistency
        # of BNode naming in query results and in graph.

        self.violationDict = {}

        # Find triples (bn ?p ?obj) and put them into violationDict[k].
        # Continue to follow obj if it's a BNode again.
        def followBNode(k, bn):
            for (s, p, obj) in self.results_graph:
                if s == bn:
                    self.violationDict[k].add((s, p, obj))
                    if isinstance(obj, BNode):
                        followBNode(k, obj)

        for (s, p, obj) in self.results_graph:
            if p == SH.result:  # SH.result's obj must be a BNode
                # New graph for the violation and bind namespaces
                self.violationDict[obj] = Graph()
                for n in self.namespaceDict:
                    self.violationDict[obj].bind(n, self.namespaceDict[n])
                # Follow the BNode
                followBNode(obj, obj)

        # find the offending triple(s) for each violation graph and add into it
        for k, violation in self.violationDict.items():
            self.__triplesForOneViolation(violation)

        return list(self.violationDict.values())

    # end of __attachOffendingTriples()

    # Load namespaces into a dictionary which is accumulative with
    # the shape graph and data graph.
    def __buildNamespaceDict(self, g):
        for (prefix, path) in g.namespaces():
            assert (prefix not in self.namespaceDict) or (
                Namespace(path) == self.namespaceDict[prefix]
            ), f"Same prefix {prefix} used for {self.namespaceDict[prefix]} and {path}"

            if prefix not in self.namespaceDict:
                self.namespaceDict[prefix] = Namespace(path)

    # Query data graph and return the list of resulting triples
    def __queryDataGraph(self, s, p, o):
        q = prepareQuery(
            "SELECT ?s ?p ?o WHERE {%s %s %s .}"
            % (s if s else "?s", p if p else "?p", o if o else "?o"),
            initNs=self.namespaceDict,
        )
        res = self.data_graph.query(q)
        assert len(res), f"Must have at lease one triple like '{s} {p} {o}'"
        return res

    # Take one contraint violation (a graph) and a sh: predicate,
    # find the object which is a node in the data graph.
    # Return the object found or None.
    def __violationPredicateObj(self, violation, predicate, mustFind=True):
        q = prepareQuery(
            f"SELECT ?s ?p ?o WHERE {{ ?s {predicate} ?o . }}",
            initNs=self.namespaceDict,
        )
        res = violation.query(q)
        if mustFind:
            assert len(res) == 1, f"Must have predicate '{predicate}'"
        if len(res):
            for (s, p, o) in res:
                return o
        return None  # Ok to miss certain predicate, such as sh:resultPath

    # Take one contraint violation (a graph) and find the potential offending
    # triples.  Return the triples in a list.
    def __triplesForOneViolation(self, violation):
        resultPath = self.__violationPredicateObj(
            violation, "sh:resultPath", mustFind=False
        )
        if resultPath:
            focusNode = self.__violationPredicateObj(violation, "sh:focusNode")
            valueNode = self.__violationPredicateObj(
                violation, "sh:value", mustFind=False
            )

            # TODO: Although we haven't seen a violation with sh:resultPath where
            # focusNode and valueNode are the same, the case should be considered.
            # The triple probably should be queried using queryDataGraph() instead
            # of assuming focusNode is the subject here.

            if valueNode:
                g = Graph()
                g.add((focusNode, resultPath, valueNode))
                violation.add((BNode(), BSH["offendingTriple"], g))
                return
            else:
                # Without valueNode, we look for constraint, such as
                # sh:class <class> and sh:minCount <number>
                cComp = self.__violationPredicateObj(
                    violation, "sh:sourceConstraintComponent"
                )
                c = cComp.split("#")[1].replace("ConstraintComponent", "")
                cPred = "sh:" + c[0].lower() + c[1:]
                cObj = f"<{self.__violationPredicateObj(violation, cPred)}>"

                g = Graph()
                g.add((focusNode, resultPath, Literal(f"{cPred} {cObj}")))
                violation.add((BNode(), BSH["offenderHint"], g))

            return
        # end of if resultPath:

        # Without sh:resultPath or sh:value in the violation. We are currently only
        # concerned with the RDFS.domain shape.
        sourceShape = self.__violationPredicateObj(violation, "sh:sourceShape")
        if sourceShape.endswith("DomainShape"):
            (bsh, shapeName) = sourceShape.split("#")

            # For a brick property xyz with RDFS.domain predicate, the shape's name
            # is bsh:xyzDomainShape.  Here we tease out brick:xyz to make the query.
            brickProp = shapeName[: -len("DomainShape")]
            path = "brick:" + brickProp
            fullPath = self.namespaceDict["brick"] + brickProp

            # The full name (http...) of the focusNode doesn't seem to work
            # in the query.  Therefore make a prefixed version for the query.
            focusNode = self.__violationPredicateObj(violation, "sh:focusNode")
            res = self.__queryDataGraph(f"<{focusNode}>", path, None)

            # Due to inherent ambiguity of this kind of shape,
            # multiple triples may be found.
            for (s, p, o) in res:
                g = Graph()
                g.add((focusNode, URIRef(fullPath), o))
                violation.add((BNode(), BSH["offendingTriple"], g))
            return
        # end of if sourceShape.endswith('DomainShape'):

        # When control reaches here, a handler is missing for the violation.

        self.log.error(
            "no triple finder for violation %s"
            % violation.serialize(format="ttl").decode("utf-8")
        )
        return

    # end of triplesForOneViolation()

    # Serialize and streamline (remove @prefix lines) a grpah and append to output
    def __appendGraph(self, msg, g):
        if msg:
            self.extraOutput += msg
        for n in self.namespaceDict:
            g.bind(n, self.namespaceDict[n])

        for b_line in g.serialize(format="ttl").splitlines():
            line = b_line.decode("utf-8")
            # skip prefix, offendingTriple and blank line
            if (
                (not line.startswith("@prefix"))
                and ("offenderHint" not in line)
                and ("offendingTriple" not in line)
                and line.strip()
            ):
                self.extraOutput += line
                self.extraOutput += "\n"

    def __appendViolation(self, msg, g):
        # first print the violation body
        self.__appendGraph(msg, g)

        # tease out the triples with offendingTriple as predicate
        tripleGraphs = []
        tripleType = None
        for (s, p, o) in g:
            if p == BSH["offendingTriple"] or p == BSH["offenderHint"]:
                tripleType = p
                tripleG = Graph()
                for (s1, p1, o1) in o:
                    tripleG.add((s1, p1, o1))
                tripleGraphs.append(tripleG)

        if len(tripleGraphs) == 0:
            self.extraOutput += "Please let us know if the contraint violation information is insufficient.\n"
            return

        if tripleType == BSH["offenderHint"]:
            self.extraOutput += "Violation hint (subject predicate cause):\n"
        elif len(tripleGraphs) == 1:
            self.extraOutput += "Offending triple:\n"
        else:
            self.extraOutput += "Potential offending triples:\n"
        for tripleG in tripleGraphs:
            self.__appendGraph(None, tripleG)

    def __getExtraOutput(self):
        self.extraOutput = f"\nAdditional info ({len(self.violationList)} constraint violations with offender hint):\n"

        # Print each violation graph, find and print the offending triple(s), too
        for g in self.violationList:
            self.__appendViolation("\nConstraint violation:\n", g)


# end of class Validator()



import logging
import itertools
import csv
import re
import os
import pkgutil
import io
import pickle
from collections import defaultdict
from .namespaces import BRICK, A, RDFS
import rdflib
from .tagmap import tagmap
import owlrl
import tarfile


class OWLRLNaiveInferenceSession:
    """
    Provides methods and an inferface for producing the deductive closure
    of a graph under OWL-RL semantics. WARNING this may take a long time
    """

    def expand(self, graph):
        """
        Applies OWLRL reasoning from the Python owlrl library to the graph

        Args:
            graph (brickschema.graph.Graph): a Graph object containing triples
        """
        owlrl.DeductiveClosure(owlrl.OWLRL_Semantics).expand(graph)


class OWLRLReasonableInferenceSession:
    """
    Provides methods and an inferface for producing the deductive closure
    of a graph under OWL-RL semantics. WARNING this may take a long time
    """

    def __init__(self):
        """
        Creates a new OWLRL Inference session
        """
        try:
            from reasonable import PyReasoner
        except ImportError:
            raise ImportError(
                "'reasonable' package not found. Install\
support for the reasonable Reasoner with 'pip install brickschema[reasonable].\
Currently only works on Linux and MacOS"
            )
        self.r = PyReasoner()

    def expand(self, graph):
        """
        Applies OWLRL reasoning from the Python reasonable library to the graph

        Args:
            graph (brickschema.graph.Graph): a Graph object containing triples
        """
        self.r.from_graph(graph)
        triples = self.r.reason()
        graph.add(*triples)


class OWLRLAllegroInferenceSession:
    """
    Provides methods and an inferface for producing the deductive closure
    of a graph under OWL-RL semantics. WARNING this may take a long time

    Uses the Allegrograph reasoning implementation
    """

    def __init__(self):
        """
        Creates a new OWLRL Inference session backed by the Allegrograph
        reasoner (https://franz.com/agraph/support/documentation/current/materializer.html).
        Requires the docker package to work; recommended method of installing
        is to use the 'allegro' option with pip:
            pip install brickschema[allegro]
        """

        try:
            import docker
        except ImportError:
            raise ImportError(
                "'docker' package not found. Install support \
for Allegro with 'pip install brickschema[allegro]"
            )

        self._client = docker.from_env(version="auto")
        containers = self._client.containers.list(all=True)
        print(f"Checking {len(containers)} containers")
        for c in containers:
            if c.name != "agraph":
                continue
            if c.status == "running":
                print("Killing running agraph")
                c.kill()
            print("Removing old agraph", c.status)
            c.remove(v=True)
            break

    def _setup_input(self, g):
        """
        Add our serialized graph to an in-memory tar file
        that we can send to Docker
        """
        g.serialize("input.ttl", format="turtle")
        tarbytes = io.BytesIO()
        tar = tarfile.open(name="out.tar", mode="w", fileobj=tarbytes)
        tar.add("input.ttl", arcname="input.ttl")
        tar.close()
        # seek to beginning so our file is not empty when docker sees it
        tarbytes.seek(0)
        return tarbytes

    def expand(self, graph):
        """
        Applies OWLRL reasoning from the Python owlrl library to the graph

        Args:
            graph (brickschema.graph.Graph): a Graph object containing triples
        """

        def check_error(res):
            exit_code, message = res
            if exit_code > 0:
                print(f"Non-zero exit code {exit_code} with message {message}")

        # setup connection to docker
        tar = self._setup_input(graph)
        # TODO: temporary name so we can have more than one running?
        agraph = self._client.containers.run(
            "franzinc/agraph:v7.0.0", name="agraph", detach=True, shm_size="1G"
        )
        if not agraph.put_archive("/tmp", tar):
            print("Could not add input.ttl to docker container")
        check_error(agraph.exec_run("chown -R agraph /tmp", user="root"))
        check_error(
            agraph.exec_run(
                "/agraph/bin/agraph-control --config /agraph/etc/agraph.cfg start",
                user="agraph",
            )
        )
        check_error(
            agraph.exec_run(
                "/agraph/bin/agload test \
/tmp/input.ttl",
                user="agraph",
            )
        )
        check_error(
            agraph.exec_run(
                "/agraph/bin/agmaterialize test \
--rule all",
                user="agraph",
            )
        )
        check_error(
            agraph.exec_run(
                "/agraph/bin/agexport -o turtle test\
 /tmp/output.ttl",
                user="agraph",
            )
        )
        bits, stat = agraph.get_archive("/tmp/output.ttl")
        with open("output.ttl.tar", "wb") as f:
            for chunk in bits:
                f.write(chunk)
        tar = tarfile.open("output.ttl.tar")
        tar.extractall()
        tar.close()

        agraph.stop()
        agraph.remove(v=True)
        graph.load_file("output.ttl")

        # cleanup
        os.remove("output.ttl")
        os.remove("output.ttl.tar")
        os.remove("input.ttl")


class VBISTagInferenceSession:
    """
    Add appropriate VBIS tag annotations to the entities inside the provided Brick model

    Algorithm:
    - get all Equipment entities in the Brick model (VBIs currently only deals w/ equip)

    Args:
        alignment_file (str): use the given Brick/VBIS alignment file. Defaults to a
                pre-packaged version.
        master_list_file (str): use the given VBIS tag master list. Defaults to a
                pre-packaged version.

    Returns:
        A VBISTagInferenceSession object
    """

    def __init__(self, alignment_file=None, master_list_file=None):
        self._alignment_file = alignment_file
        self._master_list_file = master_list_file

        from .graph import Graph

        self._graph = Graph()
        if self._alignment_file is None:
            data = pkgutil.get_data(
                __name__, "ontologies/Brick-VBIS-alignment.ttl"
            ).decode()
            self._graph.parse(source=io.StringIO(data), format="ttl")
        else:
            self._graph.load_file(self._alignment_file)

        if self._master_list_file is None:
            data = pkgutil.get_data(__name__, "ontologies/vbis-masterlist.csv").decode()
            master_list_file = io.StringIO(data)
        else:
            master_list_file = open(self._master_list_file)

        # query the graph for all VBIS patterns that are linked to Brick classes
        # Build a lookup table from the results
        self._pattern2class = defaultdict(list)
        self._class2pattern = {}
        res = self._graph.query(
            """SELECT ?class ?vbispat WHERE {
            ?shape  a   sh:NodeShape .
            ?shape  sh:targetClass  ?class .
            { ?shape  sh:property/sh:pattern ?vbispat }
            UNION
            { ?shape  sh:or/rdf:rest*/rdf:first/sh:pattern ?vbispat }
        }"""
        )
        for row in res:
            brickclass, vbispattern = row
            self._pattern2class[vbispattern].append(brickclass)
            self._class2pattern[brickclass] = vbispattern

        # Build a lookup table of VBIS pattern -> VBIS tag. The VBIS patterns
        # used as keys are from the above lookup table, so they all correspond
        # to a Brick class
        self._pattern2vbistag = defaultdict(list)
        rdr = csv.DictReader(master_list_file)
        for row in rdr:
            for pattern in self._pattern2class.keys():
                if re.match(pattern, row["VBIS Tag"]):
                    self._pattern2vbistag[pattern].append(row["VBIS Tag"])
        master_list_file.close()

    def expand(self, graph):
        """
        Args:
            graph (brickschema.graph.Graph): a Graph object containing triples
        """

        ALIGN = rdflib.Namespace(
            "https://brickschema.org/schema/1.1/Brick/alignments/vbis#"
        )
        graph += self._graph

        equip_and_shape = graph.query(
            """SELECT ?equip ?class ?shape WHERE {
             ?class rdfs:subClassOf* brick:Equipment .
             ?equip rdf:type ?class .
             ?shape sh:targetClass ?class .
        }"""
        )
        equips = set([row[0] for row in equip_and_shape])
        for equip in equips:
            rows = [row for row in equip_and_shape if row[0] == equip]
            classes = set([row[1] for row in rows])
            brickclass = self._filter_to_most_specific(graph, classes)
            applicable_vbis = self._pattern2vbistag[self._class2pattern[brickclass]]
            if len(applicable_vbis) == 1:
                graph.add((equip, ALIGN.hasVBISTag, rdflib.Literal(applicable_vbis[0])))
            elif len(applicable_vbis) > 1:
                common_pfx = _get_common_prefix(applicable_vbis)
                graph.add((equip, ALIGN.hasVBISTag, rdflib.Literal(common_pfx)))
            else:
                logging.info(f"No VBIS tags found for {equip} with type {brickclass}")

    def _filter_to_most_specific(self, graph, classlist):
        """
        Given a list of Brick classes (rdflib.URIRef), return the most specific one
        (the one that is not a superclass of the others)
        """
        candidates = {}
        for brickclass in classlist:
            sc_query = f"SELECT ?subclass WHERE {{ ?subclass rdfs:subClassOf+ <{brickclass}> }}"
            subclasses = set([x[0] for x in graph.query(sc_query)])
            # if there are NO subclasses of 'brickclass', then it is specific
            if len(subclasses) == 0:
                candidates[brickclass] = 0
                continue
            # 'subclasses' are the subclasses of 'brickclass'. If any of these appear in
            # 'classlist', then we know that 'brickclass' is not the most specific
            intersection = set(classlist).intersection(subclasses)
            if len(intersection) == 1 and brickclass in intersection:
                candidates[brickclass] = 1
            else:
                candidates[brickclass] = len(intersection)
        most_specific = None
        mincount = float("inf")
        for specific, score in candidates.items():
            if score < mincount:
                most_specific = specific
                mincount = score
        return most_specific

    def lookup_brick_class(self, vbistag):
        """
        Returns all Brick classes that are appropriate for the given VBIS tag

        Args:
            vbistag (str): the VBIS tag  that we want to retrieve Brick classes for. Pattern search
                is not supported yet
        Returns:
            brick_classes (list of rdflib.URIRef): list of the Brick classes that match the VBIS tag
        """
        if "*" in vbistag:
            raise Exception("Pattern search not supported in current release")
        classes = set()
        for pattern, brickclasses in self._pattern2class.items():
            if re.match(pattern, vbistag):
                classes.update(brickclasses)
        return list(classes)


class TagInferenceSession:
    """
    Provides methods and an interface for inferring Brick classes from
    sets of Brick tags. If you want to work with non-Brick tags, you
    will need to use a wrapper class (see HaystackInferenceSession)
    """

    def __init__(
        self,
        load_brick=True,
        rebuild_tag_lookup=False,
        approximate=False,
        brick_file=None,
    ):
        """
        Creates new Tag Inference session
        Args:
            load_brick (bool): if True, load Brick ontology into the graph
            brick_file (str): path to a Brick ttl file to use; will replace
                the internal version of Brick if provided and will treat
                'load_brick' as False
            rebuild_tag_lookup (bool): if True, rebuild the dictionary
                used for performing the inference of tags -> classes.
                By default, uses the dictionary for the packaged Brick
                version
            approximate (bool): if True, considers a more permissive set of
                possibly related classes. If False, performs exact tag mapping
        """
        from .graph import Graph

        if brick_file is not None:
            self.g = Graph(load_brick=False)
            self.g.load_file(brick_file)
        else:
            self.g = Graph(load_brick=load_brick)
        self._approximate = approximate
        if rebuild_tag_lookup:
            self._make_tag_lookup()
        else:
            # get ontology data from package
            data = pkgutil.get_data(__name__, "ontologies/taglookup.pickle")
            # TODO: move on from moving pickle to something more secure?
            self.lookup = pickle.loads(data)

    def _make_tag_lookup(self):
        """
        Builds taglookup dictionary. You shouldn't need to do this unless
        the taglookup dictionary is out of date
        """
        self.lookup = defaultdict(set)
        res = self.g.query(
            """SELECT ?class ?tag WHERE {
          ?class rdfs:subClassOf+ brick:Class.
          ?class brick:hasAssociatedTag ?tag .
          ?tag rdf:type brick:Tag
        }"""
        )
        class2tag = defaultdict(set)
        for (cname, tag) in res:
            cname = cname.split("#")[1]
            tag = tag.split("#")[1]
            class2tag[cname].add(tag)
        for cname, tagset in class2tag.items():
            self.lookup[tuple(sorted(tagset))].add(cname)
        pickle.dump(self.lookup, open("taglookup.pickle", "wb"))

    def _is_point(self, classname):
        return (
            len(
                self.g.query(
                    f"SELECT ?x WHERE {{ \
            brick:{classname} rdfs:subClassOf* brick:Point . \
            brick:{classname} a ?x }}"
                )
            )
            > 0
        )

    def _is_equip(self, classname):
        return (
            len(
                self.g.query(
                    f"SELECT ?x WHERE {{ \
            brick:{classname} rdfs:subClassOf* brick:Equipment . \
            brick:{classname} a ?x }}"
                )
            )
            > 0
        )

    def _translate_tags(self, tags):
        """"""
        output_tags = []
        for tag in tags:
            tag = tag.lower()
            if tag not in tagmap:
                output_tags.append(tag)
                continue
            output_tags.extend(tagmap[tag])
        return set(output_tags)

    def lookup_tagset(self, tagset):
        """
        Returns the Brick classes and tagsets that are supersets OR
        subsets of the given tagsets

        Args:
            tagset (list of str): a list of tags
        """
        s = set(map(_to_tag_case, tagset))
        if self._approximate:
            s.add("Point")
            withpoint = [
                (klass, set(tagset))
                for tagset, klass in self.lookup.items()
                if s.issuperset(set(tagset)) or s.issubset(set(tagset))
            ]
            s.remove("Point")
            s.add("Equipment")
            withequip = [
                (klass, set(tagset))
                for tagset, klass in self.lookup.items()
                if s.issuperset(set(tagset)) or s.issubset(set(tagset))
            ]
            return withpoint + withequip

        return [
            (klass, set(tagset))
            for tagset, klass in self.lookup.items()
            if s == set(tagset)
        ]

    def most_likely_tagsets(self, orig_s, num=-1):
        """
        Returns the list of likely classes for a given set of tags,
        as well as the list of tags that were 'leftover', i.e. not
        used in the inference of a class

        Args:
            tagset (list of str): a list of tags
            num (int): number of likely tagsets to be returned; -1 returns all

        Returns:
            results (tuple): a 2-element tuple containing (1)
            most_likely_classes (list of str): list of Brick classes
            and (2) leftover (set of str): list of tags not used

        """
        s = set(map(_to_tag_case, orig_s))
        tagsets = self.lookup_tagset(s)
        if len(tagsets) == 0:
            # no tags
            return [], orig_s
        # find the highest number of tags that overlap
        most_overlap = max(map(lambda x: len(s.intersection(x[1])), tagsets))

        # return the class with the fewest tags >= the overlap size
        candidates = list(
            filter(lambda x: len(s.intersection(x[1])) == most_overlap, tagsets)
        )

        # When calculating the minimum difference, we calculate it form the
        # perspective of the candidate tagsets because they will have more tags
        # We want to find the tag set(s) who has the fewest tags over what was
        # provided
        min_difference = min(map(lambda x: len(x[1].difference(s)), candidates))
        most_likely = list(
            filter(lambda x: len(x[1].difference(s)) == min_difference, candidates)
        )

        leftover = s.difference(most_likely[0][1])
        most_likely_classes = list(set([list(x[0])[0] for x in most_likely]))
        # return most likely classes (list) and leftover tags
        # (what of 'orig_s' wasn't used)
        if num < 0:
            return most_likely_classes, leftover
        else:
            return most_likely_classes[:num], leftover

    def expand(self, graph):
        """
        Infers the Brick class for entities with tags; tags are indicated
        by the `brick:hasTag` relationship.
        Args:
            graph (brickschema.graph.Graph): a Graph object containing triples
        """
        for triple in self.g:
            graph.add(triple)
        entity_tags = defaultdict(set)
        res = graph.query(
            """SELECT ?ent ?tag WHERE {
            ?ent brick:hasTag ?tag
        }"""
        )
        for ent, tag in res:
            entity_tags[ent].add(tag)
        for entity, tagset in entity_tags.items():
            tagset = list(map(lambda x: x.split("#")[-1], tagset))
            lookup = self.lookup_tagset(tagset)
            if len(lookup) == 0:
                continue
            klasses = list(lookup[0][0])
            graph.add((entity, A, BRICK[klasses[0]]))


class HaystackInferenceSession(TagInferenceSession):
    """
    Wraps TagInferenceSession to provide inference of a Brick model
    from a Haystack model. The haystack model is expected to be encoded
    as a dictionary with the keys "cols" and "rows"; I believe this is
    a standard Haystack JSON export.
    """

    def __init__(self, namespace):
        """
        Creates a new HaystackInferenceSession that infers entities into
        the given namespace
        Args:
            namespace (str): namespace into which the inferred Brick entities
                             are deposited. Should be a valid URI
        """
        super(HaystackInferenceSession, self).__init__(
            approximate=True, load_brick=True
        )
        self._generated_triples = []
        self._BLDG = rdflib.Namespace(namespace)
        self._filters = [
            lambda x: not x.startswith("his"),
            lambda x: not x.endswith("Ref"),
            lambda x: not x.startswith("cur"),
            lambda x: x != ("disMacro"),
            lambda x: x != "navName",
            lambda x: x != "tz",
            lambda x: x != "mod",
            lambda x: x != "id",
        ]
        self._point_tags = [
            "point",
            "sensor",
            "command",
            "setpoint",
            "alarm",
            "status",
            "parameter",
            "limit",
        ]

    def infer_entity(self, tagset, identifier=None, equip_ref=None):
        """
        Produces the Brick triples representing the given Haystack tag set

        Args:
            tagset (list of str): a list of tags representing a Haystack entity
            equip_ref (str): reference to an equipment if one exists

        Keyword Args:
            identifier (str): if provided, use this identifier for the entity,
            otherwise, generate a random string.
        """
        triples = []
        infer_results = []
        if identifier is None:
            raise Exception("PROVIDE IDENTIFIER")

        # handle Site
        if "site" in tagset and "equip" not in tagset and "point" not in tagset:
            triples.append((self._BLDG[identifier.replace(" ", "_")], A, BRICK.Site))
            return triples, [(identifier, list(tagset), [BRICK.Site])]

        # take into account 'equipref' to avoid unnecessarily inventing equips
        if equip_ref is not None:
            equip_entity_id = equip_ref
            inferred_equip_classes = []
        else:
            non_point_tags = set(tagset).difference(self._point_tags)
            non_point_tags.add("equip")
            inferred_equip_classes, leftover_equip = self.most_likely_tagsets(
                non_point_tags
            )
            inferred_equip_classes = [
                c for c in inferred_equip_classes if self._is_equip(c)
            ]
            equip_entity_id = identifier.replace(" ", "_") + "_equip"

        # choose first class for now
        point_entity_id = identifier.replace(" ", "_") + "_point"

        # check if this is a point; if so, infer what it is
        if set(tagset).intersection(self._point_tags):
            tagset = set(tagset).difference(set(["equip"]))
            inferred_point_classes, leftover_points = self.most_likely_tagsets(tagset)
            inferred_point_classes = [
                c for c in inferred_point_classes if self._is_point(c)
            ]
            if len(inferred_point_classes) > 0:
                triples.append(
                    (self._BLDG[point_entity_id], A, BRICK[inferred_point_classes[0]])
                )
                triples.append(
                    (
                        self._BLDG[point_entity_id],
                        RDFS.label,
                        rdflib.Literal(identifier),
                    )
                )
                infer_results.append((identifier, list(tagset), inferred_point_classes))

        if len(inferred_equip_classes) > 0:
            triples.append(
                (self._BLDG[equip_entity_id], A, BRICK[inferred_equip_classes[0]])
            )
            triples.append(
                (
                    self._BLDG[equip_entity_id],
                    BRICK.hasPoint,
                    self._BLDG[point_entity_id],
                )
            )
            triples.append(
                (
                    self._BLDG[equip_entity_id],
                    RDFS.label,
                    rdflib.Literal(identifier + " equip"),
                )
            )
            triples.append(
                (
                    self._BLDG[point_entity_id],
                    RDFS.label,
                    rdflib.Literal(identifier + " point"),
                )
            )
            infer_results.append((identifier, list(tagset), inferred_equip_classes))
        return triples, infer_results

    def _translate_tags(self, haystack_tags):
        """"""
        output_tags = []
        for tag in haystack_tags:
            tag = tag.lower()
            if tag not in tagmap:
                output_tags.append(tag)
                continue
            output_tags.extend(tagmap[tag])
        return set(output_tags)

    def infer_model(self, model):
        """
        Produces the inferred Brick model from the given Haystack model
        Args:
            model (dict): a Haystack model
        Returns:
            graph (brickschema.graph.Graph): a Graph object containing the
                inferred triples in addition to the regular graph
        """

        from .graph import Graph

        entities = model["rows"]
        # index the entities by their ID field
        entities = {e["id"].replace('"', ""): {"tags": e} for e in entities}
        # TODO: add e['dis'] for a descriptive label?
        brickgraph = Graph(load_brick=False)

        # marker tag pass
        for entity_id, entity in entities.items():
            marker_tags = {
                k for k, v in entity["tags"].items() if v == "m:" or v == "M"
            }
            for f in self._filters:
                marker_tags = list(filter(f, marker_tags))
            # translate tags
            entity_tagset = list(self._translate_tags(marker_tags))

            equip_ref = entity["tags"].get("equipRef")
            # infer tags for single entity
            triples, _ = self.infer_entity(
                entity_tagset, identifier=entity_id, equip_ref=equip_ref
            )
            brickgraph.add(*triples)
            self._generated_triples.extend(triples)

        # take a pass through for relationships
        for entity_id, entity in entities.items():
            relships = {k: v for k, v in entity["tags"].items() if k.endswith("Ref")}
            # equip_entity_id = entity_id.replace(' ', '_') + '_equip'
            point_entity_id = entity_id.replace(" ", "_") + "_point"
            if "equipRef" not in relships:
                continue
            reffed_equip = (
                relships["equipRef"].replace(" ", "_").replace('"', "") + "_equip"
            )
            if self._BLDG[point_entity_id] in brickgraph.nodes:
                triple = (
                    self._BLDG[reffed_equip],
                    BRICK.hasPoint,
                    self._BLDG[point_entity_id],
                )
                brickgraph.add(triple)
                self._generated_triples.append(triple)
        return brickgraph


def _get_common_prefix(list_of_strings):
    """
    Returns the longest common prefix among the set of strings.
    Helpful for finding a VBIS tag prefix.

    Args:
        list_of_strings (list of str): list of strings
    Returns:
        pfx (str): longest common prefix
    """
    # https://stackoverflow.com/questions/6718196/determine-prefix-from-a-set-of-similar-strings
    def all_same(x):
        return all(x[0] == y for y in x)

    char_tuples = zip(*list_of_strings)
    prefix_tuples = itertools.takewhile(all_same, char_tuples)
    return "".join(x[0] for x in prefix_tuples).strip("-")


def _to_tag_case(x):
    """
    Returns the string in "tag case" where the first letter
    is capitalized

    Args:
        x (str): input string
    Returns:
        x (str): transformed string
    """
    return x[0].upper() + x[1:]



"""
ORM for Brick
"""
from collections import defaultdict
from . import namespaces as ns
from sqlalchemy import Column, String, ForeignKey, create_engine
from sqlalchemy.orm import relationship, sessionmaker
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
# TODO: brick:feeds (many-to-many), brick:hasPart


class Equipment(Base):
    """
    SQLAlchemy ORM class for BRICK.Equipment; see SQLORM class for usage
    """

    __tablename__ = "equipment"

    name = Column(String, primary_key=True)
    type = Column(String)

    points = relationship("Point", back_populates="equipment")

    location_id = Column(ForeignKey("location.name"))
    location = relationship("Location", back_populates="equipment")


class Point(Base):
    """
    SQLAlchemy ORM class for BRICK.Point; see SQLORM class for usage
    """

    __tablename__ = "point"

    name = Column(String, primary_key=True)
    type = Column(String)
    equipment_id = Column(ForeignKey("equipment.name"))
    equipment = relationship("Equipment", back_populates="points")

    location_id = Column(ForeignKey("location.name"))
    location = relationship("Location", back_populates="points")


class Location(Base):
    """
    SQLAlchemy ORM class for BRICK.Location; see SQLORM class for usage
    """

    __tablename__ = "location"

    name = Column(String, primary_key=True)
    type = Column(String)
    equipment = relationship("Equipment", back_populates="location")
    points = relationship("Point", back_populates="location")


class SQLORM:
    """
    A SQLAlchemy-based ORM for Brick models.

    Currently, the ORM models Locations, Points and Equipment and the
    basic relationships between them.
    """

    def __init__(self, graph, connection_string="sqlite://brick_orm.db"):
        """
        Creates a new ORM instance over the given Graph using SQLAlchemy.
        The ORM does not capture *all* information expressed in a Brick model,
        but can be easily extended over time to capture more information.

        Args:
            graph (brickschema.Graph): a Brick schema graph containing
                instances we want to interact with. **Note**: this graph
                should not have any inference applied to it (RDFS or otherwise)
            connection_string (str): a database URL telling SQLAlchemy how to
                connect to the database that is backing the ORM. See
                [SQLAlchemy's documentation on database URLs](https://docs.sqlalchemy.org/en/13/core/engines.html#database-urls)
        """
        self._graph = graph
        self._engine = create_engine(connection_string)
        Base.metadata.create_all(self._engine)

        # the SQLAlchemy session; use for queries, etc
        self.session = sessionmaker(bind=self._engine)()

        # populate the database

        # get all equipment
        res = self._graph.query(
            """SELECT ?equip ?type WHERE {
            ?equip  rdf:type/rdfs:subClassOf* brick:Equipment .
            ?equip  rdf:type ?type
        }"""
        )
        for (equip_name, equip_type) in res:
            equip = Equipment(name=equip_name, type=equip_type)
            self.session.merge(equip)

        # get all points of equipment
        res = self._graph.query(
            """SELECT ?point ?type ?equip WHERE {
            ?point  rdf:type/rdfs:subClassOf* brick:Point .
            ?point  rdf:type ?type .
            {
                ?point  brick:isPointOf ?equip .
            } UNION {
                ?equip  brick:hasPoint ?point .
            }
        }"""
        )
        for (point_name, point_type, equip_name) in res:
            point = Point(name=point_name, type=point_type, equipment_id=equip_name)
            self.session.merge(point)

        # get all locations
        res = self._graph.query(
            """SELECT ?location ?type WHERE {
            ?location  rdf:type/rdfs:subClassOf* brick:Location .
            ?location  rdf:type ?type .
        }"""
        )
        for (loc_name, loc_type) in res:
            loc = Location(name=loc_name, type=loc_type)
            self.session.merge(loc)

        # get all locations of equipment
        res = self._graph.query(
            """SELECT ?location ?type ?equip WHERE {
            ?equip  rdf:type/rdfs:subClassOf* brick:Equipment .
            ?location  rdf:type/rdfs:subClassOf* brick:Location .
            ?location  rdf:type ?type .
            {
                ?location  brick:isLocationOf ?equip .
            } UNION {
                ?equip  brick:hasLocation ?location .
            }
        }"""
        )
        for (loc_name, loc_type, equip_name) in res:
            # get existing equip object
            equip = (
                self.session.query(Equipment).filter(Equipment.name == equip_name).one()
            )
            # get existing Location
            loc = self.session.query(Location).filter(Location.name == loc_name).one()
            self.session.merge(loc)
            loc.equipment.append(equip)
            self.session.merge(loc)

        self.session.commit()


class _DynamicORM:
    """
    TODO: under construction
    Object-oriented interface to Brick models. We turn triples into objects
    as follows:

    1. Equipment with points: all instances of equipment
    """

    def __init__(self, graph):
        """
        Creates a new ORM instance over the given Graph

        Args:
            graph (brickschema.Graph): a Brick schema graph containing
                instances we want to interact with. **Note**: this graph
                should not have any inference applied to it (RDFS or otherwise)
        """
        self.graph = graph

        # construct initial Equipment class
        self.equipment_classes = {
            "Equipment": type(
                "Equipment",
                (),
                {
                    "URI": ns.BRICK["Equipment"],
                    "classname": "Equipment",
                    "name": None,
                    "__repr__": _brick_repr,
                    "points": [],
                },
            ),
        }
        # construct subclasses recursively
        self._build_subclasses(
            self.equipment_classes["Equipment"], self.equipment_classes, set()
        )

        self.point_classes = {
            "Point": type(
                "Point",
                (),
                {"URI": ns.BRICK["Point"], "classname": "Point", "name": None},
            ),
        }
        self._build_subclasses(self.point_classes["Point"], self.point_classes, set())

        # get equipment instances
        self.instances = {}
        for name, klass in self.equipment_classes.items():
            res = self.graph.query(
                f"""SELECT ?inst ?point ?pointtype WHERE {{
                ?inst a <{klass.URI}> .
                ?inst brick:hasPoint ?point .
                ?point a ?pointtype
            }}"""
            )
            for (inst, point, pointtype) in res:
                inst_name = inst.split("#")[-1]
                class_inst = self.instances.get(inst_name, klass())
                pointtype = pointtype.split("#")[-1]
                if pointtype not in self.point_classes:
                    print(f"No instances of {pointtype} for {klass}")
                    continue
                point_class = self.point_classes[pointtype]
                point_inst = point_class()
                point_inst.name = point
                class_inst.name = inst_name
                class_inst.points.append(point_inst)
                self.instances[inst_name] = class_inst

    def _build_subclasses(self, rootclass, dest, visited=None):
        if rootclass.URI in visited:
            return
        visited.add(rootclass.URI)
        res = self.graph.query(
            f"""SELECT ?class WHERE {{
            ?class rdfs:subClassOf <{rootclass.URI}>
        }}"""
        )
        for row in res:
            class_uri = row[0]
            name = class_uri.split("#")[-1]
            klass = type(
                name,
                (rootclass,),
                {
                    "URI": class_uri,
                    "classname": name,
                    "__repr__": _brick_repr,
                    "name": None,
                    "points": [],
                },
            )
            dest[name] = klass
            self._build_subclasses(klass, dest, visited=visited)


def _brick_repr(self):
    return f"<BRICK {self.classname}: {self.name}>"



"""
Brickschema web module. This embeds a Flask webserver which provides a local web server with:
- SPARQL interpreter + query result visualization
- buttons to perform inference

TODO:
- implement https://www.w3.org/TR/sparql11-protocol/ on /query
"""
from flask import Flask, request, json, jsonify, redirect
from rdflib.plugins.sparql.results.jsonresults import JSONResultSerializer
import pkgutil
import io


class Server:
    def __init__(self, graph):
        self.graph = graph
        self.app = Flask(__name__, static_url_path="/static")

        self.app.route("/query", methods=["GET", "POST"])(self.query)
        self.app.route("/reason/<profile>", methods=["POST"])(self.apply_reasoning)
        self.app.route("/", methods=["GET"])(self.home)

    def query(self):
        if request.method == "GET":
            query = request.args.get("query")
        elif (
            request.method == "POST"
            and request.content_type == "application/x-www-form-urlencoded"
        ):
            query = request.form.get("query")
        elif (
            request.method == "POST"
            and request.content_type == "application/sparql-query"
        ):
            print("SPARQL", request.form.keys())
            query = request.get_data()
        print(query)
        results = self.graph.query(query)
        json_results = io.StringIO()
        JSONResultSerializer(results).serialize(json_results)
        return jsonify(json.loads(json_results.getvalue()))

    def home(self):
        return pkgutil.get_data(__name__, "web/index.html").decode()

    def apply_reasoning(self, profile):
        self.graph.expand(profile)
        return jsonify(len(self.graph))

    def start(self, address):
        assert len(address.split(":")) == 2
        host, port = address.split(":")
        self.app.run(host="localhost", port="8080")



"""
The `namespaces` module provides pointers to standard Brick namespaces
and related ontology namespaces
wrapper class and convenience methods for a Brick graph
"""
from rdflib import Namespace

BRICK = Namespace("https://brickschema.org/schema/1.1/Brick#")
TAG = Namespace("https://brickschema.org/schema/1.1/BrickTag#")
BSH = Namespace("https://brickschema.org/schema/1.1/BrickShape#")
OWL = Namespace("http://www.w3.org/2002/07/owl#")
RDF = Namespace("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
RDFS = Namespace("http://www.w3.org/2000/01/rdf-schema#")
SKOS = Namespace("http://www.w3.org/2004/02/skos/core#")
SH = Namespace(f"http://www.w3.org/ns/shacl#")

A = RDF.type


def bind_prefixes(graph):
    """
    Associate common prefixes with the graph
    """
    graph.bind("rdf", RDF)
    graph.bind("owl", OWL)
    graph.bind("rdfs", RDFS)
    graph.bind("skos", SKOS)
    graph.bind("brick", BRICK)
    graph.bind("tag", TAG)
    graph.bind("bsh", BSH)
    graph.bind("sh", SH)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['brickschema', 'brickschema.bin']

package_data = \
{'': ['*'], 'brickschema': ['ontologies/*', 'web/*']}

install_requires = \
['Flask>=1.1.2,<2.0.0',
 'owlrl>=5.2,<6.0',
 'pyshacl>=0.12.1,<0.13.0',
 'pytest>=6.2,<7.0',
 'rdflib>=5.0,<6.0',
 'requests>=2.24.0,<3.0.0',
 'sqlalchemy>=1.3,<2.0']

extras_require = \
{'allegro': ['docker>=4.1,<5.0'], 'reasonable': ['reasonable>=0.1.22,<0.2.0']}

entry_points = \
{'console_scripts': ['brick_validate = brickschema.bin.brick_validate:main']}

setup_kwargs = {
    'name': 'brickschema',
    'version': '0.2.0',
    'description': 'A library for working with the Brick ontology for buildings (brickschema.org)',
    'long_description': '# Brick Ontology Python package\n\n![Build](https://github.com/BrickSchema/py-brickschema/workflows/Build/badge.svg)\n[![Documentation Status](https://readthedocs.org/projects/brickschema/badge/?version=latest)](https://brickschema.readthedocs.io/en/latest/?badge=latest)\n\nDocumentation available at [readthedocs](https://brickschema.readthedocs.io/en/latest/)\n\n## Installation\n\nThe `brickschema` package requires Python >= 3.6. It can be installed with `pip`:\n\n```\npip install brickschema\n```\n\nThe `brickschema` package offers several installation configuration options for reasoning.\nThe default bundled [OWLRL](https://pypi.org/project/owlrl/) reasoner delivers correct results, but exhibits poor performance on large or complex ontologies (we have observed minutes to hours) due to its bruteforce implementation.\n\nThe [Allegro reasoner](https://franz.com/agraph/support/documentation/current/materializer.html) has better performance and implements enough of the OWLRL profile to be useful. We execute Allegrograph in a Docker container, which requires the `docker` package. To install support for the Allegrograph reasoner, use\n\n```\npip install brickschema[allegro]\n```\n\nThe [reasonable Reasoner](https://github.com/gtfierro/reasonable) offers even better performance than the Allegro reasoner, but is currently only packaged for Linux and MacOS platforms. To install support for the reasonable Reasoner, use\n\n```\npip install brickschema[reasonable]\n```\n\n## Quickstart\n\nThe main `Graph` object is just a subclass of the excellent [RDFlib Graph](https://rdflib.readthedocs.io/en/stable/) library, so all features on `rdflib.Graph` will also work here.\n\nBrief overview of the main features of the `brickschema` package:\n\n```python\nimport brickschema\n\n# creates a new rdflib.Graph with a recent version of the Brick ontology\n# preloaded.\ng = brickschema.Graph(load_brick=True)\n# OR use the absolute latest Brick:\n# g = brickschema.Graph(load_brick_nightly=True)\n# OR create from an existing model\n# g = brickschema.Graph(load_brick=True).from_haystack(...)\n\n# load in data files from your file system\ng.load_file("mbuilding.ttl")\n# ...or by URL (using rdflib)\ng.parse("https://brickschema.org/ttl/soda_brick.ttl", format="ttl")\n\n# perform reasoning on the graph (edits in-place)\ng.expand(profile="owlrl")\ng.expand(profile="tag") # infers Brick classes from Brick tags\n\n# validate your Brick graph against built-in shapes (or add your own)\nvalid, _, resultsText = g.validate()\nif not valid:\n    print("Graph is not valid!")\n    print(resultsText)\n\n# perform SPARQL queries on the graph\nres = g.query("""SELECT ?afs ?afsp ?vav WHERE  {\n    ?afs    a       brick:Air_Flow_Sensor .\n    ?afsp   a       brick:Air_Flow_Setpoint .\n    ?afs    brick:isPointOf ?vav .\n    ?afsp   brick:isPointOf ?vav .\n    ?vav    a   brick:VAV\n}""")\nfor row in res:\n    print(row)\n\n# start a blocking web server with an interface for performing\n# reasoning + querying functions\ng.serve("localhost:8080")\n# now visit in http://localhost:8080\n```\n\n## Features\n\n## Inference\n\n`brickschema` makes it easier to employ reasoning on your graphs. Simply call the `expand` method on the Graph object with one of the following profiles:\n- `"rdfs"`: RDFS reasoning\n- `"owlrl"`: OWL-RL reasoning (using 1 of 3 implementations below)\n- `"vbis"`: add VBIS tags to Brick entities\n- `"tag"`: infer Brick classes from Brick tags\n\n\n```python\nfrom brickschema import Graph\n\ng = Graph(load_brick=True)\ng.load_file("test.ttl")\ng.expand(profile="owlrl")\nprint(f"Inferred graph has {len(g)} triples")\n```\n\n\nThe package will automatically use the fastest available reasoning implementation for your system:\n\n- `reasonable` (fastest, Linux-only for now): `pip install brickschema[reasonable]`\n- `Allegro` (next-fastest, requires Docker): `pip install brickschema[allegro]`\n- OWLRL (default, native Python implementation): `pip install brickschema`\n\nTo use a specific reasoner, specify `"reasonable"`, `"allegrograph"` or `"owlrl"` as the value for the `backend` argument to `graph.expand`.\n\n## Haystack Translation\n\n`brickschema` can produce a Brick model from a JSON export of a Haystack model.\nThen you can use this package as follows:\n\n```python\nimport json\nfrom brickschema import Graph\nmodel = json.load(open("haystack-export.json"))\ng = Graph(load_brick=True).from_haystack("http://project-haystack.org/carytown#", model)\npoints = g.query("""SELECT ?point ?type WHERE {\n    ?point rdf:type/rdfs:subClassOf* brick:Point .\n    ?point rdf:type ?type\n}""")\nprint(points)\n```\n\n## VBIS Translation\n\n`brickschema` can add [VBIS](https://vbis.com.au/) tags to a Brick model easily\n\n```python\nfrom brickschema import Graph\ng = Graph(load_brick=True)\ng.load_file("mybuilding.ttl")\ng.expand(profile="vbis")\n\nvbis_tags = g.query("""SELECT ?equip ?vbistag WHERE {\n    ?equip  <https://brickschema.org/schema/1.1/Brick/alignments/vbis#hasVBISTag> ?vbistag\n}""")\n```\n\n## Web-based Interaction\n\n`brickschema` now supports interacting with a Graph object in a web browser. Executing `g.serve(<http address>)` on a graph object from your Python script or interpreter will start a webserver listening (by default) at http://localhost:8080 . This uses [Yasgui](https://yasgui.triply.cc/) to provide a simple web interface supporting SPARQL queries and inference.\n\n\n## Brick model validation\n\nThe module utilizes the [pySHACL](https://github.com/RDFLib/pySHACL) package to validate a building ontology against the Brick Schema, its default constraints (shapes) and user provided shapes.\n\n```python\nfrom brickschema import Graph\n\ng = Graph(load_brick=True)\ng.load_file(\'myBuilding.ttl\')\nvalid, _, _ = g.validate()\nprint(f"Graph is valid? {valid}")\n\n# validating using externally-defined shapes\nexternal = Graph()\nexternal.load_file("other_shapes.ttl")\nvalid, _, _ = g.validate(shape_graphs=[external])\nprint(f"Graph is valid? {valid}")\n```\n\nThe module provides a command\n`brick_validate` similar to the `pyshacl` command.  The following command is functionally\nequivalent to the code above.\n```bash\nbrick_validate myBuilding.ttl -s other_shapes.ttl\n```\n\n## Development\n\nBrick requires Python >= 3.6. We use [pre-commit hooks](https://pre-commit.com/) to automatically run code formatters and style checkers when you commit.\n\nUse [Poetry](https://python-poetry.org/docs/) to manage packaging and dependencies. After installing poetry, install dependencies with:\n\n```bash\n# -D flag installs development dependencies\npoetry install -D\n```\n\nEnter the development environment with the following command (this is analogous to activating a virtual environment.\n\n```bash\npoetry shell\n```\n\nOn first setup, make sure to install the pre-commit hooks for running the formatting and linting tools:\n\n```bash\n# from within the environment; e.g. after running \'poetry shell\'\npre-commit install\n```\n\nRun tests to make sure build is not broken\n\n```bash\n# from within the environment; e.g. after running \'poetry shell\'\nmake test\n```\n\n### Docs\n\nDocs are written in reStructured Text. Make sure that you add your package requirements to `docs/requirements.txt`\n',
    'author': 'Gabe Fierro',
    'author_email': 'gtfierro@cs.berkeley.edu',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://brickschema.org',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
